#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <NeoPixelBus.h>

#include "FS.h"  // Spiff file system

#include <wind_misc_utils.h>
#include <wind_webserver.h>

const int NeoPixelsPin = D4;
#define PIXELS_COUNT 144
#define PIXEL_MAX_LEFT 71
#define PIXEL_MIN_RIGHT 72
#define PIXEL_DELAY 25  // milliseconds between pixel changes

const char WiFiAPPSK[] = "wankel";  // This seems to be ignored.

WindWebServer* webserver;

ADC_MODE(ADC_TOUT);
 
// Note: ESP-01 will ignore pin # here, and use GPIO2, 
//    According to https://github.com/Makuna/NeoPixelBus/blob/master/examples/NeoPixelTest/NeoPixelTest.ino
NeoPixelBus<NeoGrbFeature, NeoEsp8266Uart800KbpsMethod> pixels(PIXELS_COUNT, NeoPixelsPin);

// TODO: Make Blade class and move color variables, these methods, and NeoPixels stuff there.
int red;
int green; 
int blue;
  
void setPixelRGB(int pixel, int red, int green, int blue) {
  pixels.SetPixelColor(pixel, RgbColor(red,green,blue));
  pixels.Show();
}

void setAllPixelsRGB(int red, int green, int blue) {
  for(int i = 0; i < PIXELS_COUNT; i++) {
    pixels.SetPixelColor(i, RgbColor(red, green, blue));
  }
  pixels.Show();
}

void scanAllPixelsDownRGB(int red, int green, int blue) {
  for(int i = 0; i < PIXELS_COUNT / 2; i++) {
    pixels.SetPixelColor(PIXEL_MAX_LEFT - i, RgbColor(red,green,blue));
    pixels.SetPixelColor(PIXEL_MIN_RIGHT + i, RgbColor(red,green,blue));
    pixels.Show();
    delay(PIXEL_DELAY);
  }
}

void scanAllPixelsUpRGB(int red, int green, int blue) {
   for(int i = 0; i < PIXELS_COUNT / 2; i++) {
    pixels.SetPixelColor(i, RgbColor(red,green,blue));
    pixels.SetPixelColor(PIXELS_COUNT - i, RgbColor(red,green,blue));
    pixels.Show();
    delay(PIXEL_DELAY);
  }
}

void BladeTurnOff() {
  scanAllPixelsDownRGB(0, 0, 0);
  red = 0;
  green = 0;
  blue = 0;
}

void BladeTurnOn(int r, int g, int b) {
  scanAllPixelsUpRGB(r, g, b);
  red = r;
  green = g;
  blue = b;
}

void BladeTransitionTo(int new_red, int new_green, int new_blue) {
  normalizeRGBtoMax(125, &new_red, &new_green, &new_blue);
  Serial.printf("Normalized colors: r=%d g=%d b=%d.\n", new_red, new_green, new_blue);
  if (red != new_red || green != new_green || blue != new_blue) {
    if (red || green || blue) {
      BladeTurnOff();
    }
    if (new_red || new_green || new_blue) {
      BladeTurnOn(new_red, new_green, new_blue);
    }
  }
}

void setup(void) {
  Serial.begin(115200);
  Serial.println("In setup.");

  setupSpiffs();
  
  delay(200);  // TODO: Not sure needed.

  // Initialize NeoPixels color.
  pixels.Begin();
  red = 255;
  green = 0;
  blue = 0;

  setupWiFiAccessPoint();
  
  // Start the server
  webserver = new WindWebServer(80);
  BladeTurnOn(red, green, blue); 
}

void setupSpiffs() {
  bool spiffs_ok = SPIFFS.begin();
  Serial.printf("SPIFFS.begin()=%d.\n", spiffs_ok);

  Serial.printf("Scanning SPIFFS directory.\n", spiffs_ok);
  Dir dir = SPIFFS.openDir("/");
  while (dir.next()) {
      Serial.print(dir.fileName());
      File f = dir.openFile("r");
      Serial.printf(".  size=%d.\n", f.size());
  }
}

void setupWiFiAccessPoint() {
  WiFi.mode(WIFI_AP);

  // Do a little work to get a unique-ish name. Append the
  // last two bytes of the MAC (HEX'd) to "Thing-":
  // TODO: Figure out why SSID isn't always named as expected.
  uint8_t mac[WL_MAC_ADDR_LENGTH];
  WiFi.softAPmacAddress(mac);
  String macID = String(mac[WL_MAC_ADDR_LENGTH - 2], HEX) +
                 String(mac[WL_MAC_ADDR_LENGTH - 1], HEX);
  macID.toUpperCase();
  String AP_NameString = "ESP8266 Thing " + macID;

  char AP_NameChar[AP_NameString.length() + 1];
  memset(AP_NameChar, 0, AP_NameString.length() + 1);

  for (int i=0; i<AP_NameString.length(); i++)
    AP_NameChar[i] = AP_NameString.charAt(i);

  WiFi.softAP(AP_NameChar, WiFiAPPSK);
  Serial.println("Creating hotspot at SSID=" + AP_NameString);
}

void HandleClientRequest() {
  Serial.println("debug hack: In HandleClientRequest().");
  
  int new_red = red;
  int new_green = green;
  int new_blue = blue;
  String file_name("/index.html");
  
  // Read the client's request: POST or GET.
  String request = webserver->ReadLine();
  Serial.println("First line of request= " + request);
  if (request.startsWith("POST ")) {
    request = webserver->ReadPost();
    Serial.println("Read POST content=" + request + ".");
    parseRGB(request, '{', ' ', &new_red, &new_green, &new_blue);
  } else {
    // Parse GET URL params.
    file_name = parseFileName(request);
    Serial.println("Read GET for file=" + file_name + ".");
    if (request.indexOf('?', 0) > 0) {
      parseRGB(request, '?', '&', &new_red, &new_green, &new_blue);
    }
  }

  String output("Parsed values: ");  
  Serial.println(output + "red=" + new_red + " green=" + new_green + " blue=" + new_blue);
  BladeTransitionTo(new_red, new_green, new_blue);

  // Return the response
  webserver->ReturnFile(file_name.c_str());
  
  delay(10);
  Serial.println("Client disconnected");
  Serial.println("");
}


void loop() {   
  // Check for a Wifi client.
  webserver->CheckForClient();

  // Handle requests to change the LED settings.
  if (webserver->ClientAvailable()) {
    Serial.println("Client available");
    HandleClientRequest();
    return;
  }

  // Give ESP8266 Watchdog Timer a chance to succeed.
  delay(50);
}
